import json
import logging
import os

import pytest

from mediawiki import mediawiki as mediawiki_module

from hive.common.testing import test_config_dir  # noqa: F401
from hive.mediawiki import HiveWiki


def test_login(test_config, mock_requests, caplog):
    auth_headers = {
        "Authorization": "Basic aHR0cC11c2VybmFtZTpodHRwLXBhc3N3b3Jk",
    }
    csrf_token = "9ed1499d99c0c34c73faa07157b3b6075b427365+\\"
    mock_requests.expect(
        MockRequest(
            method="GET",
            url="https://hive.host/path/to/api",
            headers=auth_headers,
            params={
                "action": "query",
                "meta": "tokens",
                "type": "login",
                "format": "json"
            },
        ),
        MockResponse(
            json={
                "batchcomplete": "",
                "query": {
                    "tokens": {
                        "logintoken": csrf_token,
                    },
                },
            },
        ))
    mock_requests.expect(
        MockRequest(
            method="POST",
            url="https://hive.host/path/to/api",
            headers=auth_headers,
            data={
                "action": "login",
                "lgname": "wiki-username",
                "lgpassword": "wiki-password",
                "lgtoken": csrf_token,
                "format": "json"
            },
        ),
        MockResponse(
            json={
                "login": {
                    "lguserid": 21,
                    "result": "Success",
                    "lgusername": "William"
                }
            },
        ))
    mock_requests.expect(
        MockRequest(
            method="GET",
            url="https://hive.host/path/to/api",
            headers=auth_headers,
            params={
                "action": "query",
                "meta": "siteinfo",
                "siprop": "extensions|general",
                "format": "json"
            },
        ),
        MockResponse(
            json={
                "query": {
                    "general": {
                        "generator": "MediaWiki 1.35.10",
                        "server": "https://hive.host",
                    },
                    "extensions": [
                        {"name": name}
                        for name in [
                                "MinervaNeue", "Vector", "Cite",
                                "Scribunto", "TemplateStyles",
                                "ParserFunctions", "SyntaxHighlight",
                                "Interwiki", "MobileFrontend",
                        ]
                    ],
                },
            },
        ))

    with caplog.at_level(logging.DEBUG):
        wiki = HiveWiki()
    assert wiki.logged_in
    assert wiki.api_version == "1.35.10"
    assert wiki.base_url == "https://hive.host"
    assert wiki.extensions == [
        "Cite", "Interwiki", "MinervaNeue", "MobileFrontend",
        "ParserFunctions", "Scribunto", "SyntaxHighlight",
        "TemplateStyles", "Vector"]


@pytest.fixture
def test_config(test_config_dir):  # noqa: F811
    filename = os.path.join(test_config_dir, "mediawiki.json")
    config = {
        "url": "https://hive.host/path/to/api",
        "http_auth": {
            "username": "http-username",
            "password": "http-password",
        },
        "username": "wiki-username",
        "password": "wiki-password",
    }
    with open(filename, "w") as fp:
        json.dump({"mediawiki": config}, fp)
    yield config


@pytest.fixture
def mock_requests(monkeypatch):
    module = MockRequests()
    with monkeypatch.context() as m:
        m.setattr(mediawiki_module, "requests", module)
        yield module
    assert not module._expectations


class MockRequests:
    def __init__(self):
        self._expectations = []

    def expect(self, want_request, mock_response):
        self._expectations.append((want_request, mock_response))

    def Session(self):
        return MockSession(self._expectations)


class MockSession:
    def __init__(self, expectations):
        self._expectations = expectations
        self.headers = {}
        self.auth = None
        self._is_open = True

    def close(self):
        assert self._is_open
        self._is_open = False

    def get(self, *args, **kwargs):
        return self._handle("GET", *args, **kwargs)

    def post(self, *args, **kwargs):
        return self._handle("POST", *args, **kwargs)

    def _handle(self, *args, **kwargs):
        assert self._is_open

        got_request = MockRequest(*args, **kwargs)
        if self.auth:
            got_request = self.auth(got_request)

        timeout = got_request.timeout
        assert timeout
        assert timeout > 0
        assert timeout < 300

        want_request, mock_response = self._expectations.pop(0)
        assert got_request == want_request
        return mock_response


class MockRequest:
    def __init__(self, method, url, timeout=None, headers=None, **kwargs):
        self.method = method
        self.url = url
        self.headers = headers or {}
        self.timeout = timeout
        self.kwargs = kwargs

    def __eq__(self, other):
        return not (self != other)

    def __ne__(self, other):
        for attr in ("method", "url", "headers", "kwargs"):
            a, b = (getattr(ob, attr) for ob in (self, other))
            if a == b:
                continue
            print(f"{attr}: {a!r} != {b!r}")
            return True
        return False


class MockResponse:
    def __init__(self, json):
        self._json = json

    def json(self):
        return self._json



from __future__ import annotations

import weakref

from datetime import timedelta
from typing import Any, Callable, Optional

from mediawiki import (
    MediaWiki as PyMediaWiki,
    MediaWikiPage as PyMediaWikiPage,
)
from requests import PreparedRequest
from requests.auth import AuthBase, HTTPBasicAuth

from hive.common import read_config
from hive.common.units import MINUTES

from .__version__ import __version__

__url__ = "https://github.com/gbenson/hive"


class HiveWiki(PyMediaWiki):
    DEFAULT_CONFIG_KEY = "mediawiki"
    MAX_REQUEST_TIMEOUT = 10 * MINUTES
    MIN_REQUEST_INTERVAL = timedelta(seconds=0.1)

    def __init__(self, **kwargs):
        config_sect = self.DEFAULT_CONFIG_KEY
        config_key = kwargs.pop("config_key", config_sect)

        if config_key:
            config = read_config(config_key)
            config = config.get(config_sect, {})
            config.update(kwargs)
            kwargs = config

        http_auth = kwargs.pop("http_auth", None)
        if isinstance(http_auth, dict):
            http_auth = HTTPBasicAuth(**http_auth)
        self._http_auth = http_auth

        if "rate_limit" not in kwargs:
            kwargs["rate_limit"] = True
        if "rate_limit_wait" not in kwargs:
            kwargs["rate_limit_wait"] = self.MIN_REQUEST_INTERVAL

        self._need_user_agent = not bool(kwargs.get("user_agent"))
        super().__init__(**kwargs)
        assert not self._need_user_agent

    def _reset_session(self):
        if self._need_user_agent:
            self._need_user_agent = False
            self.user_agent = \
                f"Hivetool/{__version__} ({self.user_agent}; +{__url__})"
            return

        self.__workaround_rate_limit_uninit()
        self._hive_validate_config()
        super()._reset_session()
        session = self._session
        if not (auth := session.auth):
            auth = self._http_auth
        session.auth = HiveAuthenticator(weakref.proxy(self), auth)

    def __workaround_rate_limit_uninit(self):
        """PyMediaWiki 0.7.4 indirectly calls _reset_session()
        before configuring rate limiting."""  # XXX remove
        try:
            _ = self.rate_limit
        except AttributeError:
            self._rate_limit = True
        try:
            _ = self.rate_limit_min_wait
        except AttributeError:
            self._min_wait = self.MIN_REQUEST_INTERVAL

    def _hive_validate_config(self) -> None:
        """Raise a ValueError if the current configuration is unsafe.
        """
        timeout = self.timeout
        if not timeout or timeout < 0 or timeout > self.MAX_REQUEST_TIMEOUT:
            raise ValueError(f"timeout: {timeout!r}")

        if not self.rate_limit:
            raise ValueError(f"rate_limit: {self.rate_limit!r}")

        interval = self.rate_limit_min_wait
        if interval < self.MIN_REQUEST_INTERVAL:
            raise ValueError(f"rate_limit_min_wait: {interval!r}")

        if not self.verify_ssl:
            raise ValueError(f"verify_ssl: {self.verify_ssl!r}")

    def page(self, *args, **kwargs) -> HiveWikiPage:
        return HiveWikiPage(super().page(*args, **kwargs))


class HiveAuthenticator(AuthBase):
    def __init__(self, wiki, auth=None):
        self._wiki = wiki
        self._auth = auth

    def __call__(self, r: PreparedRequest) -> PreparedRequest:
        self._wiki._hive_validate_config()

        if not self._auth:
            return r

        if not r.url.startswith("https://"):
            raise ValueError(r.url)

        session = self._wiki._session
        if not session.verify:
            raise ValueError(f"session.verify: {session.verify!r}")

        return self._auth(r)


class HiveWikiPage:
    def __init__(self, wrapped: PyMediaWikiPage):
        self._hive_wrapped = wrapped

    def __getattr__(self, attr):
        return getattr(self._hive_wrapped, attr)

    def append(self, wikitext: str) -> None:
        # XXX this could race, need to send a revision id or timestamp
        page_wikitext = self.wikitext
        if "\n" not in page_wikitext[len(page_wikitext.rstrip()):]:
            wikitext = f"\n{wikitext.rstrip()}"
        self._append(wikitext)

    def _append(self, wikitext: str) -> None:
        params = {
            "action": "query",
            "meta": "tokens",
            "format": "json",
        }
        resp = self.mediawiki._get_response(params)
        token = resp["query"]["tokens"]["csrftoken"]

        params = {
            "action": "edit",
            "appendtext": wikitext,
            "format": "json",
            "token": token,
        }

        if self.pageid:
            params["pageid"] = self.pageid
        elif self.title:
            params["title"] = self.title

        resp = self.mediawiki._post_response(params)
        self.mediawiki._check_error_response(resp, "append")

        # XXX probably need to invalidate more
        self._hive_wrapped._wikitext = None
